/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.syncclient.sps.common;

import java.io.*;
import java.util.Properties;
import java.util.Hashtable;
import java.util.Vector;
import java.util.StringTokenizer;

import org.apache.xerces.parsers.*;
import org.w3c.dom.*;
import org.w3c.dom.traversal.*;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.EntityReference;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import java.text.MessageFormat;
import java.net.URL;

import com.funambol.syncclient.common.StringTools;

/**
 *
 * This class generate java source by parsing template with XML descriptor. It can be used in two
 * manners: creating an instance and calling <i>getContent()</i> or as a
 * standalone program. In the latter case the syntax is the following:<br>
 *
 * <pre>
 *  java SPSGenerator {javaTokenTemplate1FileName} {javaTokenTemplate2FileName}
                         {javaClassTemplateFileName} {xmlDescriptorFileName} {dataStoreManagerFileName}
 * </pre>
 *
 * The compiled source is saved in file with {source_java} name.
 *
 * DTD of XML file
 *
 * <!DOCTYPE SPS-definition[
 * <!ELEMENT SPS-definition (header, datastore+)>
 * <!ELEMENT application-creator-id (#PCDATA)>
 * <!ELEMENT application-datastore-type (#PCDATA)>
 * <!ELEMENT application-description (#PCDATA)>
 * <!ELEMENT application-display-name (#PCDATA)>
 * <!ELEMENT application-support-mail (#PCDATA)>
 * <!ELEMENT application-support-url (#PCDATA)>
 * <!ELEMENT application-name (#PCDATA)>
 * <!ELEMENT application-author (#PCDATA)>
 * <!ELEMENT content-id (#PCDATA)>
 * <!ELEMENT attributes (function, size, display-name)>
 * <!ELEMENT datastore (name, display-name, default-sync, record, store-volume)>
 * <!ELEMENT default-sync (#PCDATA)>
 * <!ELEMENT display-name (#PCDATA)>
 * <!ELEMENT application-version (#PCDATA)>
 * <!ELEMENT field (name, type, attributes)>
 * <!ELEMENT function (#PCDATA)>
 * <!ELEMENT header (store-manager-package, application-uri, application-version, application-author, application-creator-id, application-datastore-type, content-id, application-display-name, application-description, application-support-url, application-support-mail)>
 * <!ELEMENT name (#PCDATA)>
 * <!ELEMENT record (field+)>
 * <!ELEMENT size (#PCDATA)>
 * <!ELEMENT store-manager-package (#PCDATA)>
 * <!ELEMENT store-volume (#PCDATA)>
 * <!ELEMENT type (#PCDATA)>
 * ]>
 *
 * $Id: SPSGenerator.java,v 1.4 2007/01/11 19:21:41 nichele Exp $
 */
public class SPSGenerator {

    //---------------------------------------------------------------- Constants

    /** source generetor file properties */
    public static final String SPSGENERATOR_PROPERTIES = "/sps.properties";

    /**sps sync source class name */
    public static final String SPS_SYNC_SOURCE_CLASS_NAME = "com.funambol.syncclient.spds.source.SPSSyncSource";

    /**sps sync item type */
    public static final String SPS_SYNC_ITEM_TYPE = "plain/record";

    public static final String IMPL_PALM     = "palm"    ;
    public static final String IMPL_POCKETPC = "pocketpc";
    public static final String IMPL_J2ME     = "j2me"    ;
    public static final String IMPL_JDBC     = "jdbc"    ;

    public static final String VALID_TARGETS = ','
                                             + IMPL_PALM + ','
                                             + IMPL_POCKETPC + ','
                                             + IMPL_J2ME + ','
                                             + IMPL_JDBC
                                             + ','
                                             ;

    // ------------------------------------------------------------ Private data

    private String javaTokenTemplate1FileName  = null;
    private String javaTokenTemplate2FileName  = null;
    private String javaTokenTemplate3FileName  = null;
    private String javaClassTemplateFileName   = null;

    private String xmlDescriptorFileName    = null;
    private String dataStoreManagerFileName = null;
    private String dataStoreManagerPackage  = null;

    private String implementation = null;

    private String dataStoreManagerTokenTemplate1 = null;
    private String dataStoreManagerTokenTemplate2 = null;
    private String dataStoreManagerTokenTemplate3 = null;
    private String dataStoreManagerClassTemplate  = null;

    private String dataStoreManager = null;

    private Vector htFieldsAttributesVector         = new Vector();
    private Vector htFieldsAttributesPutValueVector = new Vector();
    private Vector fieldsMetadataVector             = new Vector();
    private Vector dataStoreNameVector              = new Vector();
    private Vector recordSizeVector                 = new Vector();

    private int dataStoreCount = 0;

    //------------------------------------------------------------- Constructors

     /**
     * @param javaTemplate the string to be logged
     * @param init if true the existing content will be overwritten
     *
     */
    SPSGenerator (String xmlDescriptorFileName, String implementation) throws IOException {

            this.xmlDescriptorFileName = xmlDescriptorFileName;
            this.implementation = implementation;

            try {
                loadResources();
               } catch (IOException e) {
                throw new IOException( "Error in the directory structure ("
                + e.getClass().getName()
                + "): "
                + e.getMessage()
                );
            }
    }

    //----------------------------------------------------------- Private methods


    /**
     * Parsing xml descriptor and compose string to populate java template
     **/
    private void parsingXML() throws Exception {

        try {

            Document document;
            DOMParser parser;

            Class c;

            c = this.getClass();

            String dataStoreName = null;
            String fieldName     = null;
            String fieldType     = null;

            String fieldMetadata  = null;
            String fieldsMetadata = null;

            String htFieldAttributes  = null;
            String htFieldsAttributes = null;

            String htFieldAttributesPutValue  = null;
            String htFieldsAttributesPutValue = null;

            int recordCount = 0;
            int fieldCount  = 0;

            parser = new DOMParser();
            parser.setFeature("http://apache.org/xml/features/dom/defer-node-expansion", true);
            parser.setFeature("http://apache.org/xml/features/continue-after-fatal-error", true);

            parser.parse(xmlDescriptorFileName);

            document = parser.getDocument();

            NodeList figlia=document.getChildNodes();

            Node tempa;

            for (int r=0; r < figlia.getLength(); r++) {

                tempa = figlia.item(r);

                    if(tempa.getNodeName().equals("SPS-definition")) {

                        NodeList figli0 = tempa.getChildNodes();

                    Node temp0;

                    for (int m=0; m < figli0.getLength(); m++) {

                        temp0 = figli0.item(m);

                        if (temp0.getNodeName().equals("header")) {

                            NodeList figli1 = temp0.getChildNodes();

                            Node temp1;

                            for (int i=0; i < figli1.getLength(); i++) {

                                temp1 = figli1.item(i);

                                if (temp1.getNodeName().equals("store-manager-package")) {

                                    dataStoreManagerPackage = temp1.getChildNodes().item(0).getNodeValue();

                                    dataStoreManagerPackage = filterTagValue (dataStoreManagerPackage);

                                } else if (temp1.getNodeName().equals("application-name")) {
                                    dataStoreManagerFileName =
                                        createClassName(filterTagValue(temp1.getChildNodes().item(0).getNodeValue()));
                                }
                            }
                        }


                        if (temp0.getNodeName().equals("datastore")) {

                            NodeList figli1 = temp0.getChildNodes();

                            Node temp1;

                            for (int i=0; i < figli1.getLength(); i++) {

                                temp1 = figli1.item(i);

                                if (temp1.getNodeName().equals("name")) {
                                    dataStoreName=temp1.getChildNodes().item(0).getNodeValue();
                                }

                                if (temp1.getNodeName().equals("record")) {

                                    NodeList figli2 = temp1.getChildNodes();

                                    Node temp2;

                                    for (int j=0; j<figli2.getLength(); j++) {

                                        temp2 = figli2.item(j);


                                        if (temp2.getNodeName().equals("field")) {

                                            recordCount++;

                                            NodeList figli3 = temp2.getChildNodes();

                                            Node temp3;

                                            for (int k=0; k< figli3.getLength(); k++) {

                                                temp3 = figli3.item(k);

                                                if (temp3.getNodeName().equals("name")) {
                                                    fieldName = temp3.getChildNodes().item(0).getNodeValue();
                                                }

                                                if (temp3.getNodeName().equals("type")) {
                                                    fieldType = temp3.getChildNodes().item(0).getNodeValue();
                                                }

                                                if (temp3.getNodeName().equals("attributes")) {

                                                    htFieldAttributes = "Hashtable fieldAttributes" + fieldCount+" = new Hashtable();";

                                                    fieldMetadata = "fieldMetadata[" + fieldCount + "] = new FieldMetadataImpl(\"" + fieldName +"\", \"" + fieldType +"\", fieldAttributes" + fieldCount + ");";

                                                    if (fieldsMetadata == null) {
                                                        htFieldsAttributes = htFieldAttributes;
                                                        fieldsMetadata     = fieldMetadata;
                                                    } else {
                                                        htFieldsAttributes = htFieldsAttributes + "\n" + "        " + htFieldAttributes;
                                                        fieldsMetadata = fieldsMetadata + "\n" + "            " + fieldMetadata;
                                                    }

                                                    NodeList figli4 = temp3.getChildNodes();

                                                    Node temp4;

                                                    for (int l=0; l< figli4.getLength(); l++) {

                                                        temp4 = figli4.item(l);

                                                        if (temp4.getNodeName().length() > 0 && temp4.getChildNodes().item(0) != null) {

                                                            htFieldAttributesPutValue = "fieldAttributes" + fieldCount+".put(\"" + temp4.getNodeName() +"\", \""+ temp4.getChildNodes().item(0).getNodeValue() +"\");";

                                                            if (htFieldsAttributesPutValue == null) {
                                                                htFieldsAttributesPutValue = htFieldAttributesPutValue;
                                                            } else {
                                                                htFieldsAttributesPutValue = htFieldsAttributesPutValue + "\n" + "        " + htFieldAttributesPutValue;
                                                            }

                                                        }

                                                    }

                                                    fieldCount++;

                                                }

                                            }

                                        }

                                    }

                                }

                            }

                        }

                        if (temp0.getNodeName().equals("datastore")) {

                            dataStoreNameVector.addElement(dataStoreName);
                            recordSizeVector.addElement(String.valueOf(fieldCount));
                            htFieldsAttributesVector.addElement(htFieldsAttributes);
                            htFieldsAttributesPutValueVector.addElement(htFieldsAttributesPutValue);
                            fieldsMetadataVector.addElement(fieldsMetadata);


                            htFieldsAttributes = null;
                            htFieldsAttributesPutValue = null;
                            fieldsMetadata = null;

                            fieldCount = 0;
                            dataStoreCount++;

                        }

                    }

                }

            }

        } catch (Exception e) {
            System.out.println("Error in parsing operation.");
            throw e;
        }

    }



    /**
     * Populate java template
     * @param device implementation
     */
    private void compileSource(String implementation) {

        String dataStoreManagerTmp1 = null;
        String dataStoreManagerTmp2 = null;
        String dataStoreManagerTmp3 = null;

        String dataStoreManagerTmpA = null;
        String dataStoreManagerTmpB = null;
        String dataStoreManagerTmpC = null;


        String classPrefix = "";
        if (implementation.equalsIgnoreCase(IMPL_PALM)) {
            classPrefix = "Palm";
        } else if (implementation.equalsIgnoreCase(IMPL_POCKETPC)) {
            classPrefix = "PPC";
        } else if (implementation.equalsIgnoreCase(IMPL_J2ME)) {
            classPrefix = "J2ME";
        } else if (implementation.equalsIgnoreCase(IMPL_JDBC)) {
            classPrefix = "JDBC";
        }
        for (int i=0; i < dataStoreCount; i++) {
            dataStoreManagerTmpA = MessageFormat.format(dataStoreManagerTokenTemplate1, new Object[] {dataStoreNameVector.elementAt(i),
                                                                                        htFieldsAttributesVector.elementAt(i),
                                                                                        htFieldsAttributesPutValueVector.elementAt(i),
                                                                                        recordSizeVector.elementAt(i),
                                                                                        fieldsMetadataVector.elementAt(i),
                                                                                        classPrefix});

            dataStoreManagerTmpB = MessageFormat.format(dataStoreManagerTokenTemplate2, new Object[] {dataStoreNameVector.elementAt(i)});

            dataStoreManagerTmpC = MessageFormat.format(dataStoreManagerTokenTemplate3, new Object[] {String.valueOf(i), dataStoreNameVector.elementAt(i)});

            if (dataStoreManagerTmp1 == null) {
                dataStoreManagerTmp1 = dataStoreManagerTmpA;
            } else {
                dataStoreManagerTmp1 = dataStoreManagerTmp1 + "\n" + "\n" + dataStoreManagerTmpA;
            }


            if (dataStoreManagerTmp2 == null) {
                dataStoreManagerTmp2 = dataStoreManagerTmpB;
            } else {
                dataStoreManagerTmp2 = dataStoreManagerTmp2 + "\n" + "\n" + dataStoreManagerTmpB;
            }

            if (dataStoreManagerTmp3 == null) {
                dataStoreManagerTmp3 = dataStoreManagerTmpC;
            } else {
                dataStoreManagerTmp3 = dataStoreManagerTmp3 + "\n" + "\n" + dataStoreManagerTmpC;
            }

        }


        dataStoreManager = MessageFormat.format(dataStoreManagerClassTemplate, new Object[] {dataStoreManagerPackage, implementation, dataStoreManagerFileName, dataStoreManagerTmp1, dataStoreManagerTmp2,
                                                                                             String.valueOf(dataStoreCount), dataStoreManagerTmp3});


    }



    /**
     * Load java template
     */
    private void loadResources() throws IOException {

        Class c;

        InputStream is = null;
        Properties prop = new Properties();

        c = this.getClass();

        is = c.getResourceAsStream(SPSGENERATOR_PROPERTIES);

        prop.load(is);

        is.close();

        javaTokenTemplate1FileName = prop.getProperty("javaTokenTemplate1FileName");
        javaTokenTemplate2FileName = prop.getProperty("javaTokenTemplate2FileName");
        javaTokenTemplate3FileName = prop.getProperty("javaTokenTemplate3FileName");
        javaClassTemplateFileName  = prop.getProperty("javaClassTemplateFileName");
        dataStoreManagerFileName   = prop.getProperty("dataStoreManagerFileName");

        dataStoreManagerTokenTemplate1 = read(c.getResourceAsStream(javaTokenTemplate1FileName));
        dataStoreManagerTokenTemplate2 = read(c.getResourceAsStream(javaTokenTemplate2FileName));
        dataStoreManagerTokenTemplate3 = read(c.getResourceAsStream(javaTokenTemplate3FileName));
        dataStoreManagerClassTemplate  = read(c.getResourceAsStream(javaClassTemplateFileName));


    }



    /**
     * Save java source
     * @param implementation device implementation
     */
    private void saveSource(String implementation)
    throws IOException {
        new File(implementation).mkdir();

        String fileName =
            implementation + File.separator + dataStoreManagerFileName + ".java";

        FileOutputStream fo = new FileOutputStream(fileName);
        fo.write(dataStoreManager.getBytes());
    }


    /**
     * Save properties file
     * with key, value
     */
    private void saveProperties()  {

        String content = null;

        for (int i=0; i < dataStoreNameVector.size(); i++)

        {

            content = "sourceClass=" + SPS_SYNC_SOURCE_CLASS_NAME + "\n" +
                      "storeManager=" + this.xmlDescriptorFileName.substring(0, this.xmlDescriptorFileName.indexOf(".")).substring(0,1).toUpperCase() + this.xmlDescriptorFileName.substring(0, this.xmlDescriptorFileName.indexOf(".")).substring(1).toLowerCase() + "\n" +
                      "name=" + (String) dataStoreNameVector.elementAt(i) + "\n" +
                      "type=" + SPS_SYNC_ITEM_TYPE + "\n" +
                      "sourceURI=" + ((String) dataStoreNameVector.elementAt(i)).toLowerCase();

            try {

                FileOutputStream fo = new FileOutputStream(((String) dataStoreNameVector.elementAt(i)).toLowerCase() + ".properties");
                fo.write(content.getBytes());

            } catch(FileNotFoundException e1) {
            } catch (IOException e2) {
            }

        }

    }



    /**
     * Reads the content of the given input stream.
     *
     * @param is the input stream
     *
     **/
    private String read(InputStream is) throws IOException {
        StringBuffer sb = new StringBuffer();

        try {
            byte[] buf = new byte[1024];

            int nbyte = -1;
            while ((nbyte = is.read(buf)) >= 0) {
                sb.append(new String(buf, 0, nbyte));
            }
        } finally {
            is.close();
        }

        return sb.toString();
    }



    //-------------------------------------------------------------------- Main

    /**
     * Main
     */
    public static void main (String[] args) throws Exception {

        try {

            SPSGenerator sGenerator = null;

            checkSyntax(args);

            sGenerator = new SPSGenerator(args[0], args[1]);

            sGenerator.parsingXML();

            sGenerator.compileSource(sGenerator.implementation);

            sGenerator.saveSource(sGenerator.implementation);

            //this values are generated by client,
            //and saved by Device Management
            //sGenerator.saveProperties();

            System.out.println("Source generation OK");

        } catch (Exception e) {
            System.out.println("Source generation interruped.");
            e.printStackTrace();
        }

    }


    /**
     * Check main parameters
     *
     * @param args main parameters
     */
    private static void checkSyntax(String[] args) throws Exception {
       if (args.length != 2 || (VALID_TARGETS.indexOf(',' + args[1] + ',')<0)) {
            String msg = "Syntax error.\n"
            + "\n\tSyntax: "
            + SPSGenerator.class.getName()
            + "\n\txmlDescriptorFileName: name of xml descriptor file [e.g. ADF.xml]"
            + "\n\timplementation: device profile implementation [" + VALID_TARGETS + "]\n"
            ;
            System.out.println(msg);
            throw new Exception();
        }
    }



    /**
     * erase newline and init / end space from tag value
     *
     * @param tagValue value to filter
     * @return filtered tagValue
     */
    private String filterTagValue(String tagValue) {

        while (tagValue.indexOf("\n") != -1) {
            tagValue = tagValue.substring(0, tagValue.indexOf("\n")) + tagValue.substring(tagValue.indexOf("\n") + 1);
        }

        return tagValue.trim();
    }

    /**
     * Creates a class name replacing all special caracheters with an '_' (underscore)
     * and appending DataStore.
     *
     * @param s input string
     *
     * @return the class name created as described above.
     */
    private String createClassName(String s) {
        return StringTools.replaceSpecial(s) + "StoreManager";
    }


}