/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.syncclient.sps.pocketpc;

import java.io.IOException;
import java.util.Date;
import java.util.Hashtable;
import java.util.Vector;

import com.funambol.syncclient.sps.common.DataAccessException;
import com.funambol.syncclient.sps.common.DataStore;
import com.funambol.syncclient.sps.common.Record;
import com.funambol.syncclient.sps.common.RecordFilter;
import com.funambol.syncclient.sps.common.RecordMetadata;
import com.funambol.syncclient.spds.SyncRecordFunction;

import com.funambol.syncclient.sps.common.util.StaticDataHelper;

/**
 * This DataStore implementation makes use of JNI to access the ActiveSync
 * RAPI subsystem to read/write the hendheld object store.
 *
 * @version $Id: PPCDataStore.java,v 1.4 2007/01/11 19:21:41 nichele Exp $
 */
public class PPCDataStore extends DataStore {

    // --------------------------------------------------------------- Constants

    public final String PROPERTY_DATA_STORE_VOLUME = "dataStoreVolume";

    // ------------------------------------------------------------- PrivateData

    /**
     * Elements that must be deleted at modifications committing.
     */
    private Vector deleted = null;

    /**
     * Elements that must be updated at modifications committing.
     */
    private Vector updated = null;

    //-------------------------------------------------------------- Costructors

    public PPCDataStore(String dataStoreName, RecordMetadata recordMetadata) {
        super(dataStoreName, recordMetadata);
    }

    //----------------------------------------------------------- Public methods

    /**
     * create record
     *
     * @param key
     * @return record
     **/
    public Record newRecord(String key) {
        return new Record(this, key);
    }

    /**
     * store record
     * @param record
     * @return stored Record
     * @throws DataAccessException
     **/
    public Record storeRecord(Record record)
    throws DataAccessException {
        updated.addElement(record);

        return record;
    }

    /**
     * delete record
     * @param record
     * @throws DataAccessException
     **/
    public void deleteRecord(Record record)
    throws DataAccessException {
        deleted.addElement(record);
    }

    /**
     * read record
     * @param record
     * @throws DataAccessException
     **/
    public Record readRecord(Record record)
    throws DataAccessException {
        RapiDB db = null;
        String key = record.getKey();

        try {
            db = new RapiDB(getVolume(), getDataStoreName());

            return objectsToRecord(key, db.readRecord(key));
        } finally {
            if (db != null) {
                db.closeDB(); db = null;
            }
        }
    }

    /**
     * return all records of device stored in memory
     *
     * @return find records
     **/
    public Vector findAllRecords()
    throws DataAccessException {
        Vector records = new Vector();
        RapiDB db = null;
        Object[][] data;
        int keyPos = SyncRecordFunction.getKeyFieldPosition(getRecordMetadata());
        int statusPos = SyncRecordFunction.getModificationTypeFieldPosition(getRecordMetadata());

        try {
            db = new RapiDB(getVolume(), getDataStoreName());

            //
            // If the database does not exist, just return the empty vector
            //
            if (!db.openDB(keyPos, statusPos, keyPos)) {
                db = null;
                return records;
            }

            data = db.getAllRecords();

            if (data != null) {
                for (int i=0; i<data.length; ++i) {
                    records.addElement(objectsToRecord((String)data[i][keyPos], data[i]));
                }
            }
        } finally {
            if (db != null) {
                db.closeDB(); db = null;
            }
        }

        return records;
    }


    /**
     * return a Vector of Records
     * found by state
     *
     * @param state state of record
     * @param since last timestamp
     * @return find records
     * @throws DataAccessException
     **/
    public Vector findRecords(char state, Date since)
    throws DataAccessException {
        RapiDB db = null;
        Object[] data;
        Vector records = new Vector();
        int keyPos = SyncRecordFunction.getKeyFieldPosition(getRecordMetadata());
        int statusPos = SyncRecordFunction.getModificationTypeFieldPosition(getRecordMetadata());

        try {
            db = new RapiDB(getVolume(), getDataStoreName());

            if (!db.openDB(keyPos, statusPos, statusPos)){
                db = null;
                return records;
            }

            while ((data = db.getNextRecord(state)) != null) {
                records.addElement(objectsToRecord((String)data[keyPos], data));
            }

            return records;
        } finally {
            if (db != null) {
                db.closeDB(); db = null;
            }
        }
    }


    /**
     * return records of dataStore find by RecordFilter
     *
     * @param recordFilter filter
     * @return find records
     * @throws DataAccessException
     **/
    public Vector findRecords(RecordFilter recordFilter)
    throws DataAccessException {
        //
        // Not implemented here...
        //
        Vector all = findAllRecords();
        Vector selected = new Vector();

        Record record = null;

        int l = all.size();

        for(int i=0; i < l; i++) {

            record = (Record) all.elementAt(i);

            if(recordFilter.accept(record)) {
                selected.addElement(record);
            }
        }

        return selected;
    }

    /**
     * return next key on device
     * 1000 if no record exist
     *
     * @return key
     **/
    public synchronized long getNextKey() {
        //
        // Sleeping for one millisecond ensures that everyone gets a new value
        //
        try {
            Thread.sleep(1);
        } catch (Exception e) {};
        return System.currentTimeMillis();
    }


    /**
     * Method define start DB operathions
     * @throws DataAccessException
     */
    public void startDBOperations()
    throws DataAccessException {
        deleted = new Vector();
        updated = new Vector();
    }


    /**
     * Method define end DB operathions
     * @throws DataAccessException
     */
    public void commitDBOperations()
    throws DataAccessException {
        RapiDB db = null;
        int keyPos = SyncRecordFunction.getKeyFieldPosition(getRecordMetadata());
        int statusPos = SyncRecordFunction.getModificationTypeFieldPosition(getRecordMetadata());
        int n;
        Record record = null;

        try {
            db = new RapiDB(getVolume(), getDataStoreName());

            if (!db.openDB(keyPos, statusPos, keyPos)) {
                db.createDB(keyPos, statusPos);
                db.openDB(keyPos, statusPos, keyPos);
            }

            //
            // Deletes deleted records
            //
            n = deleted.size();
            for (int i=0; i<n; ++i) {
                db.deleteRecord(((Record)deleted.elementAt(i)).getKey());
            }

            //
            // Updates new/updataed records
            //
            n = updated.size();
            for (int i=0; i<n; ++i) {
                record = (Record)updated.elementAt(i);
                db.storeRecord(record.getKey(), recordToObjects(record));
            }

            db.closeDB(); db = null;

            //
            // Reset modification flag
            //
            Vector records = new Vector();
            records.addAll(findRecords('N', null));
            records.addAll(findRecords('U', null));

            n = records.size();
            if (n > 0) {
                db = new RapiDB(getVolume(), getDataStoreName());
                db.openDB(keyPos, statusPos, keyPos);

                for (int i=0; i<n; ++i) {
                    record = (Record)records.elementAt(i);
                    db.setRecordField(record.getKey(), statusPos, "S");
                }
            }
        } finally {
            if (db != null) {
                db.closeDB(); db = null;
            }
        }
    }

    // --------------------------------------------------------- Private methods

    private String getVolume() {
        return (String)getDataStoreProperties().get(PROPERTY_DATA_STORE_VOLUME);
    }

    /**
     * Converts a <i>Record</i> object to an <i>Object[]</i> array where each
     * element contains the value of a field.
     *
     * @param record the record - NOT NULL
     *
     * @return the <i>Object[]</i> representation of the record
     */
    private Object[] recordToObjects(Record record) {
        int l = record.getLength();
        Object[] values = new Object[l];

        for (int i=0; i<l; ++i) {
            values[i] = record.getString(i+1);
        }

        return values;
    }

    /**
     * Converts an <i>Object[]</i> into a <i>Record</i/.
     *
     * @param data the data - NOT NULL
     *
     * @return the newly created record
     **/
     private Record objectsToRecord(String key, Object[] data) {
        Record record = new Record(this, key);
        for (int i=0; i<data.length; ++i) {
            record.setField(i+1, (String) data[i]);
        }

        return record;
     }
}
